#!/usr/bin/env python
"""
Copyright (c) 2019 - present AppSeed.us
"""

import os
import sys

def main():

    try:
        
        print(' EXEC -> ' + os.path.basename(__file__)) 
        
        # Unix ErrCode
        exit(0)

    except Exception as e:

        print( 'Err: ' + str( e ) )
        exit(1)

if __name__ == '__main__':
    main()



# -*- encoding: utf-8 -*-
"""
Copyright (c) 2019 - present AppSeed.us
"""

import os, json

from django_celery_results.models import TaskResult
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver
from django.conf import settings


@receiver(pre_save, sender=TaskResult)
def pre_result_updated(sender, **kwargs):
    '''
    Before a task result is saved, the following checks and updates are carried out if task is complete:
    - Check if error key & value exist in result.result, if yes upate instance.status=`FAILURE` and instance.traceback=`logs produced during task execution`
    '''
    if 'instance' not in kwargs:
        return
    instance = kwargs.pop("instance")
    if instance.status != "STARTED":
        result = json.loads(instance.result)
        if result.get("error"):
            instance.status = "FAILURE"
            instance.traceback = result.get("logs")


@receiver(post_save, sender=TaskResult)
def post_result_updated(sender, **kwargs):
    """
    After a task is completed, we store its logs in a file.
    """
    if 'instance' not in kwargs:
        return
    instance = kwargs.pop("instance")
    if instance.status != "STARTED":
        add_log(instance)


def add_log(result: TaskResult):
    """
    Adds log file to celery logs with formatted date as name.
    :param result TaskResult: Result of executed celery task
    :rtype: None
    """
    if not result.task_name:
        return

    try:

        log_file_path = os.path.join(
                settings.CELERY_LOGS_DIR, f"{result.date_created.strftime(r'%Y%m%d-%H_%M')}-{result.task_id}-{result.status}.log")

        with open(log_file_path, "w+") as f:
            if result.status == "FAILURE":
                f.writelines([result.result, result.traceback])
            else:
                f.write(result.result)

            f.close()

    except Exception as e:
        print( 'Logging ERR: ' + str( e) )
        return
        


# -*- encoding: utf-8 -*-
"""
Copyright (c) 2019 - present AppSeed.us
"""

import os, time, json
from os import listdir
from os.path import isfile, join

from django.contrib.auth.decorators import login_required
from django.http      import HttpResponse
from django.shortcuts import redirect
from django.template  import loader
from django.conf      import settings

from celery import current_app
from .tasks import execute_script, users_in_db, get_scripts
from django_celery_results.models import TaskResult
from celery.contrib.abortable import AbortableAsyncResult
from .celery import app

@login_required(login_url="/login/")
def tasks(request):

    scripts, ErrInfo = get_scripts()
 
    context = {
               'cfgError' : ErrInfo,
               'tasks'    : get_celery_all_tasks(),
               'scripts'  : scripts}

    # django_celery_results_task_result
    task_results = TaskResult.objects.all()
    context["task_results"] = task_results

    html_template = loader.get_template('tasks/index.html')
    return HttpResponse(html_template.render(context, request)) 

def run_task(request, task_name):
    '''
    Runs a celery task
    :param request HttpRequest: Request
    :param task_name str: Name of task to execute
    :rtype: (HttpResponseRedirect | HttpResponsePermanentRedirect)
    '''
    tasks = [execute_script, users_in_db]
    _input = request.POST.get("input")
    for task in tasks:
        if task.__name__ == task_name:
            task.delay({"input": _input})
    time.sleep(1)  # Waiting for task status to update in db

    return redirect("tasks") 

def cancel_task(request, task_id):
    '''
    Cancels a celery task using its task id
    :param request HttpRequest: Request
    :param task_id str: task_id of result to cancel execution
    :rtype: (HttpResponseRedirect | HttpResponsePermanentRedirect)
    '''
    result = TaskResult.objects.get(task_id=task_id)
    abortable_result = AbortableAsyncResult(
        result.task_id, task_name=result.task_name, app=app)
    if not abortable_result.is_aborted():
        abortable_result.revoke(terminate=True)
    time.sleep(1)
    return redirect("tasks")

def get_celery_all_tasks():
    current_app.loader.import_default_modules()
    tasks = list(sorted(name for name in current_app.tasks
                        if not name.startswith('celery.')))
    tasks = [{"name": name.split(".")[-1], "script":name} for name in tasks]
    for task in tasks:
        last_task = TaskResult.objects.filter(
            task_name=task["script"]).order_by("date_created").last()
        if last_task:
            task["id"] = last_task.task_id
            task["has_result"] = True
            task["status"] = last_task.status
            task["successfull"] = last_task.status == "SUCCESS" or last_task.status == "STARTED"
            task["date_created"] = last_task.date_created
            task["date_done"] = last_task.date_done
            task["result"] = last_task.result
            if last_task.result:
                task["input"] = json.loads(last_task.result).get("input")

    return tasks

def task_output(request):
    '''
    Returns a task output 
    '''

    task_id = request.GET.get('task_id')
    task    = TaskResult.objects.get(id=task_id)

    if not task:
        return ''

    # task.result -> JSON Format
    return HttpResponse( task.result )

def task_log(request):
    '''
    Returns a task LOG file (if located on disk) 
    '''

    task_id  = request.GET.get('task_id')
    task     = TaskResult.objects.get(id=task_id)
    task_log = 'NOT FOUND'

    if not task: 
        return ''

    try: 

        # Get logs file
        all_logs = [f for f in listdir(settings.CELERY_LOGS_DIR) if isfile(join(settings.CELERY_LOGS_DIR, f))]
        
        for log in all_logs:

            # Task HASH name is saved in the log name
            if task.task_id in log:
                
                with open( os.path.join( settings.CELERY_LOGS_DIR, log) ) as f:
                    
                    # task_log -> JSON Format
                    task_log = f.readlines() 

                break    
    
    except Exception as e:

         task_log = json.dumps( { 'Error CELERY_LOGS_DIR: ' : str( e) } )

    return HttpResponse(task_log)



# -*- encoding: utf-8 -*-
"""
Copyright (c) 2019 - present AppSeed.us
"""

import os

from celery import Celery

if os.environ.get('DJANGO_SETTINGS_MODULE'):

    app = Celery('core')

    # Using a string here means the worker doesn't have to serialize
    # the configuration object to child processes.
    # - namespace='CELERY' means all celery-related configuration keys
    #   should have a `CELERY_` prefix.
    app.config_from_object('django.conf:settings', namespace='CELERY')

    # Load task modules from all registered Django apps.
    app.autodiscover_tasks()

else:
    print(' ')
    print('Celery Configuration ERROR: ') 
    print('  > "DJANGO_SETTINGS_MODULE" not set in environment (value in manage.py)')
    print('  Hint: export DJANGO_SETTINGS_MODULE=project.settings ') 
    print(' ')


# -*- encoding: utf-8 -*-
"""
Copyright (c) 2019 - present AppSeed.us
"""

from django.urls import path, re_path
from . import views

urlpatterns = [
    path('tasks', views.tasks, name="tasks"),
    path('tasks/run/<str:task_name>'  , views.run_task,    name="run-task"    ),
    path('tasks/cancel/<str:task_id>' , views.cancel_task, name="cancel-task" ),
    path('tasks/output/'              , views.task_output, name="task-output" ),
    path('tasks/log/'                 , views.task_log,    name="task-log"    ), 
]


# -*- encoding: utf-8 -*-
"""
Copyright (c) 2019 - present AppSeed.us
"""

import os, time, subprocess
from os import listdir
from os.path import isfile, join

from .celery import app
from celery.contrib.abortable import AbortableTask

from django.contrib.auth.models import User
from django.conf import settings

def get_scripts():
    """
    Returns all scripts from 'ROOT_DIR/celery_scripts'
    """
    try:
        scripts = [f for f in listdir(settings.CELERY_SCRIPTS_DIR) if isfile(join(settings.CELERY_SCRIPTS_DIR, f))]
        return scripts, None 
    except Exception as e:
        return None, 'Error CELERY_SCRIPTS_DIR: ' + str( e )    

@app.task(bind=True, base=AbortableTask)
def users_in_db(self, data: dict):
    """
    Outputs all users in DB
    :param data dict: contains data needed for task execution.
    :rtype: None
    """
    users = User.objects.all()
    return {"output": "\n".join([u.email for u in users])}

@app.task(bind=True, base=AbortableTask)
def execute_script(self, data: dict):
    """
    This task executes scripts found in settings.CELERY_SCRIPTS_DIR and logs are later generated and stored in settings.CELERY_LOGS_DIR
    :param data dict: contains data needed for task execution. Example `input` which is the script to be executed.
    :rtype: None
    """
    script = data.get("input")

    scripts, ErrInfo = get_scripts()

    if script and script in scripts:
        # Executing related script
        script_path = os.path.join(settings.CELERY_SCRIPTS_DIR, script)
        process = subprocess.Popen(
            f"python {script_path}", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        time.sleep(8)
        error = False
        if process.wait() == 0:  # If script execution successfull
            logs = process.stdout.read().decode()
        else:
            logs = process.stderr.read().decode()
            error = True

        return {"logs": logs, "input": script, "error": error, "output": ""}



# -*- encoding: utf-8 -*-
"""
Copyright (c) 2019 - present AppSeed.us
"""

from django.apps import AppConfig

class TasksConfig(AppConfig):

    """
    See https://docs.djangoproject.com/en/2.1/ref/applications/#django.apps.AppConfig
    """

    name = "django_tm"
    label = "django_tm"
    verbose_name = "Django Tasks Manager"

    def ready(self):
        from . import signals



# -*- encoding: utf-8 -*-
"""
Copyright (c) 2019 - present AppSeed.us
"""

import json

from django import template

register = template.Library()

def date_format(date):
    """
    Returns a formatted date string
    Format:  `Year-Month-Day-Hour-Minute-Second`
    Example: `2022-10-10-00-20-33`
    :param date datetime: Date object to be formatted
    :rtype: str
    """
    try:
        return date.strftime(r'%Y-%m-%d-%H-%M-%S')
    except:
        return date

register.filter("date_format", date_format)

def get_result_field(result, field: str):
    """
    Returns a field from the content of the result attibute in result 
    Example: `result.result['field']`
    :param result AbortableAsyncResult: Result object to get field from
    :param field str: Field to return from result object
    :rtype: str
    """
    result = json.loads(result.result)
    if result:
        return result.get(field)

register.filter("get_result_field", get_result_field)
